#include <fstream>
#include <cmath>
#include "TaS_Data.hh"

TaS_Input::TaS_Input(string file_name)
{  
  unsigned a, p, s, s1, s2, l, present, precedences;
  ifstream is(file_name);
  string buffer;
  char ch;

  if (!is) 
  {
    cerr << "Cannot open file " <<  file_name << endl;
    exit(1);
  }

  // Read scalars
  is >> buffer >> ch >> actors >> ch;
  is >> buffer >> ch >> scenes >> ch;
  is >> buffer >> ch >> locations >> ch;
  is >> buffer >> ch >> precedences >> ch;

  // Resize vectors
  actor_cost.resize(actors);
  location_cost.resize(locations);
  scene_duration.resize(scenes);
  scene_location.resize(scenes);
  presence.resize(actors,vector<bool>(scenes,false));
  precedence.resize(precedences);

  // Read ActorCost
  is  >> buffer >> ch >> ch;
  for (a = 0; a < actors; a++)
    is >> actor_cost[a] >> ch;
  is >> ch;

  // Read LocationCost
  is  >> buffer >> ch >> ch;
  for (l = 0; l < locations; l++)
    is >> location_cost[l] >> ch;
  is >> ch;

  // Read SceneDuration
  is  >> buffer >> ch >> ch;
  for (s = 0; s < scenes; s++)
    is >> scene_duration[s] >> ch;
  is >> ch;

  // Read SceneLocation
  is  >> buffer >> ch >> ch;
  for (s = 0; s < scenes; s++)
    is >> scene_location[s] >> ch;
  is >> ch;

  // Read Presence
  is  >> buffer >> ch >> ch >> ch;
  for (a = 0; a < actors; a++)
    for (s = 0; s < scenes; s++)
      {
        is >> present >> ch;
        presence[a][s] = present;
      }
  is >> ch >> ch;

  // Read Precedences
   is  >> buffer >> ch >> ch >> ch;
  for (p = 0; p < precedences; p++)
    {
      is >> s1 >> ch >> s2 >> ch;
      precedence[p] = make_pair(s1,s2);
    }
}

ostream& operator<<(ostream& os, const TaS_Input& in)
{
	unsigned s, a;
	for (a = 0; a < in.actors; a++)
	{
	  for (s = 0; s < in.scenes; s++)
        os << in.presence[a][s] << " ";
	  os << endl;
	}
  return os;
}

TaS_Output::TaS_Output(const TaS_Input& my_in)
  : in(my_in), scenes(in.Scenes()), position(in.Scenes())
{}

TaS_Output& TaS_Output::operator=(const TaS_Output& out)
{
  scenes = out.scenes;
  position = out.position;
  return *this;
}

ostream& operator<<(ostream& os, const TaS_Output& out)
{
  unsigned i, a, s;
  os << "[";
  for (i = 0; i < out.in.Scenes() - 1; i++)
    os << out.scenes[i] << ", ";
  os << out.scenes[i] << "]" << endl;
 
  for (i = 0; i < out.in.Scenes(); i++)
    os << " " << out.in.SceneLocation(out[i]) << " ";
  os << endl;
  for (a = 0; a < out.in.Actors(); a++)
  {
    for (i = 0; i < out.in.Scenes(); i++)
	{
		s = out[i];
	    if (out.in.Presence(a,s))
		   os << " X ";
	    else
		   os << " - ";
	}
	os << endl;
  }
  return os;
}

istream& operator>>(istream& is, TaS_Output& out)
{
  unsigned i;
  char ch;

  is >> ch;
  for (i = 0; i < out.in.Scenes(); i++)
  {
    is >> out.scenes[i] >> ch;
	out.position[out.scenes[i]] = i;
  }
  return is;
}

void TaS_Output::Reset()
{
  unsigned i;
  for (i = 0; i < in.Scenes(); i++)
  {
	  scenes[i] = i;
	  position[i] = i;
  }
}

void TaS_Output::Swap(unsigned p1, unsigned p2)
{
   unsigned s1 = scenes[p1], s2 = scenes[p2];
   swap(scenes[p1],scenes[p2]);
   position[s1] = p2;
   position[s2] = p1;
}

unsigned TaS_Output::LocationsCost() const
{
  unsigned l1, l2, i = 0, cost = 0;
  vector<bool> visited(in.Locations(), false);
  
  l1 = in.SceneLocation(scenes[i]);
  visited[l1] = true;
  for (i = 1; i < in.Scenes(); i++)
    {
      l2 = in.SceneLocation(scenes[i]);
      if (l2 != l1)
        {
          if (visited[l2])
            {
              cost += in.LocationCost(l2); 
//               cerr << "Add cost of location " << l2 << " (" << in.LocationCost(l2) 
//                    << ") for scene " << scenes[i] << " at step " << i << endl;
            }
          else 
            visited[l2] = true;	 
          l1 = l2;
        }
    } 
  return cost;
}

unsigned TaS_Output::TalentCost() const
{
	unsigned a, p, s, cost = 0;
	vector<pair<unsigned,unsigned>> on_stage(in.Actors(),make_pair(in.Scenes()-1, 0));

	// compute the first and last scene on stage for each actor
	for (s = 0; s < in.Scenes(); s++)
	{
		p = position[s];
		for (a = 0; a < in.Actors(); a++)
		{
			if (in.Presence(a,s))
			{
				if (p < on_stage[a].first)
				   on_stage[a].first = p;
			    if (p > on_stage[a].second)
				   on_stage[a].second = p;
			}
		}
	}
    //
	for (a = 0; a < in.Actors(); a++)
	{
		for (p = on_stage[a].first; p <= on_stage[a].second; p++)
		{
			s = scenes[p];
			if (!in.Presence(a,s))
			{
			  cost += in.SceneDuration(s) * in.ActorCost(a);
// 			  cerr << "Actor " << a << " unnecessarily on stage for scene " << s 
// 			       << " (in position " << p << "), cost " << in.SceneDuration(s)
// 				   << " X " << in.ActorCost(a) << endl;
			}
		}
	}			
	return cost;
}

unsigned TaS_Output::PrecedenceViolations() const
{
	unsigned p, violations = 0;
	pair<unsigned,unsigned> pp;
	for (p = 0; p < in.Precedences(); p++)
	{
	  pp = in.Precedence(p);
      if (position[pp.first] > position[pp.second])
		 violations++;
	}
	return violations;
}
